/*
 * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file ipc.c interprocess communication and shared memory helpers */

#include "para.h"
#include "error.h"
#include "ipc.h"
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

/** abort if semget() failed that many times */
#define MAX_SEMOP_RETRIES 500

/**
 * define a new mutex
 *
 * \return the identifier for the new mutex on success, -E_SEM_GET
 * on errors.
 *
 * \sa semget(2)
 */
int mutex_new(void)
{
	int ret = semget(IPC_PRIVATE, 1, IPC_CREAT | 0666);
	return ret < 0?  -E_SEM_GET : ret;
}

/**
 * destroy a mutex
 *
 * \param id the identifier of the mutex to be destroyed
 *
 * \returns Positive on success, -E_SEM_REMOVE on errors.
 *
 * \sa semctl(2)
 */
int mutex_destroy(int id)
{
	int ret = semctl(id, 0, IPC_RMID);
	return ret < 0? -E_SEM_REMOVE : 1;
}

static void para_semop(int id, struct sembuf *sops, int num)
{
	int i;

	for (i = 0; i < MAX_SEMOP_RETRIES; i++)
		if (semop(id, sops, num) >= 0)
			return;
	PARA_EMERG_LOG("semop failed %d times, aborting\n", MAX_SEMOP_RETRIES);
	exit(EXIT_FAILURE);
}

/**
 * lock the given mutex
 *
 * This function either succeeds or aborts.
 *
 * \sa semop(2), struct misc_meta_data
 */
void mutex_lock(int id)
{
	struct sembuf sops[2] = {
		{
			.sem_num = 0,
			.sem_op = 0,
			.sem_flg = SEM_UNDO
		},
		{
			.sem_num = 0,
			.sem_op = 1,
			.sem_flg = SEM_UNDO
		}
	};
	para_semop(id, sops, 2);
}

/**
 * unlock a mutex
 *
 * \param id the identifier of the mutex
 *
 * This function either succeeds or aborts.
 *
 * \sa semop(2), struct misc_meta_data
 */
void mutex_unlock(int id)
{
	struct sembuf sops[1] = {
		{
			.sem_num = 0,
			.sem_op = -1,
			.sem_flg = SEM_UNDO
		},
	};
	para_semop(id, sops, 1);
}

/**
 * create a new shared memory area of given size
 * 
 * \sa shmget(2)
 */
int shm_new(size_t size)
{
	int ret = shmget(IPC_PRIVATE, size, IPC_CREAT | IPC_EXCL | 0600);
	return ret < 0 ? -E_SHM_GET : ret;
}

/**
 * destroy the given shared memory area
 * \sa shmctl(2)
 **/
int shm_destroy(int id)
{
	struct shmid_ds shm_desc;
	int ret = shmctl(id, IPC_RMID, &shm_desc);
	return ret < 0? -E_SHM_DESTROY : ret;
}

/**
 * attach a shared memory segment
 *
 * \param id the identifier of the shared memory segment to attach
 * \param mode either ATTACH_RO (read only) or ATTACH_RW (read/write)
 * \param result points to the attached arer which to the
 *
 * \returns positive on success, -E_SHM_ATTACH on errrors.
 *
 * \sa shmat(2)
 */
int shm_attach(int id, enum shm_attach_mode mode, void **result)
{
	if (mode == ATTACH_RW) {
		*result = shmat(id, NULL, 0);
		return *result? 1 : -E_SHM_ATTACH;
	}
	*result = shmat(id, NULL, SHM_RDONLY);
	return *result? 1 : -E_SHM_ATTACH;
}

/**
 * detach a shared memory segment
 *
 * \param addr the address of the attached segment
 *
 * \returns positive on success, -E_SHM_DETACH on errors.
 *
 * \sa shmdt(2)
 */
int shm_detach(void *addr)
{
	int ret = shmdt(addr);
	return ret < 0? -E_SHM_DETACH : 1;
}
